package jp.hishidama.eval;

import java.util.*;

import jp.hishidama.eval.exp.*;
import jp.hishidama.eval.lex.Lex;
import jp.hishidama.eval.lex.LexFactory;
import jp.hishidama.eval.rule.ShareRuleValue;
import jp.hishidama.eval.var.MapVariable;
import junit.framework.TestCase;

/**
 * ÕeXg.
 * 
 * @author <a target="hishidama"
 *         href="http://www.ne.jp/asahi/hishidama/home/soft/java/eval.html">Ђ</a>
 */
public class EvalExceptionTest extends TestCase {

	Rule rule = new TestRuleFactory().getRule();

	public void testRule() {
		assertParse(EvalException.PARSE_NOT_FOUND_END_OP, "(1]");
		assertParse(EvalException.PARSE_NOT_FOUND_END_OP, "test(1");
		assertParse(EvalException.PARSE_NOT_FOUND_END_OP, "arr[1");
		assertParse(EvalException.PARSE_NOT_FOUND_END_OP, "b?1");
		assertParse(EvalException.PARSE_INVALID_OP, "* 2");
		assertParse(EvalException.PARSE_END_OF_STR, "1*");
		assertParse(EvalException.PARSE_INVALID_CHAR, "^Cv"); // Ǝ
		assertParse(EvalException.PARSE_STILL_EXIST, "a+b c");
	}

	protected void assertParse(int expected, String str) {
		Expression exp;
		try {
			exp = rule.parse(str);
		} catch (EvalException e) {
			System.out.println(e);
			// e.printStackTrace();
			assertEquals(expected, e.getErrorCode());
			return;
		}
		fail(exp.toString());
	}

	public void testSetWord() {
		AbstractExpression exp = new PlusExpression() {
			public String toString() {
				super.setWord("err");
				return super.toString();
			}
		};
		try {
			exp.toString();
		} catch (EvalException e) {
			assertEquals(EvalException.EXP_FORBIDDEN_CALL, e.getErrorCode());
			return;
		}
		fail();
	}

	public void testEval() {
		assertEvalLong(EvalException.EXP_NOT_LET, "1=2", null);
		assertEvalLong(EvalException.EXP_NOT_VARIABLE, "1[0]", null);
	}

	public void testNumber() {
		assertEvalLong(EvalException.EXP_NOT_NUMBER, "0z", null);
		assertEvalDobl(EvalException.EXP_NOT_NUMBER, "0z", null);
		assertEvalObje(EvalException.EXP_NOT_NUMBER, "0z", null);
		CharExpression c = new CharExpression(null);
		assertEvalLong(EvalException.EXP_NOT_NUMBER, c);
		assertEvalDobl(EvalException.EXP_NOT_NUMBER, c);
		StringExpression s = new StringExpression("zzz");
		assertEvalLong(EvalException.EXP_NOT_NUMBER, s);
		assertEvalDobl(EvalException.EXP_NOT_NUMBER, s);
	}

	public void testFunc() {
		assertParse(EvalException.PARSE_NOT_FUNC, "1()");
		assertEvalLong(EvalException.EXP_FUNC_CALL_ERROR, "test()", null);
		assertEvalDobl(EvalException.EXP_FUNC_CALL_ERROR, "test()", null);
		assertEvalObje(EvalException.EXP_FUNC_CALL_ERROR, "test()", null);
	}

	public void testVar() {
		Map var = new HashMap();
		var.put("str", "abc");
		assertEvalLong(EvalException.EXP_NOT_NUMBER, "str", var);
		assertEvalDobl(EvalException.EXP_NOT_NUMBER, "str", var);
		assertEvalLong(EvalException.EXP_NOT_LET_VAR, rule.parse("var=1").ae);
		assertEvalLong(EvalException.EXP_NOT_VAR_VALUE, rule.parse("var").ae);
		assertEvalDobl(EvalException.EXP_NOT_DEF_VAR, "n", var);
		assertEvalLong(EvalException.EXP_NOT_VARIABLE, rule.parse("a.x").ae);
	}

	public void testArray() {
		Map var = new HashMap();
		var.put("str", new String[] { "abc" });
		var.put("arr", new long[1]);
		assertEvalLong(EvalException.EXP_NOT_ARR_VALUE, "str[0]", var);
		assertEvalDobl(EvalException.EXP_NOT_ARR_VALUE, "str[0]", var);
		assertEvalLong(EvalException.EXP_NOT_DEF_OBJ, "b[0]", var);
		assertEvalLong(EvalException.EXP_NOT_ARR_VALUE, "arr[-1]", var);
		assertEvalLong(EvalException.EXP_NOT_DEF_OBJ, "b[0]=1", var);
		assertEvalLong(EvalException.EXP_NOT_LET_ARR, "str[0]=1", var);
	}

	public void testField() {
		Map var = new HashMap();
		var.put("f", new FieldTestA(1));
		assertEvalLong(EvalException.EXP_NOT_NUMBER, "f.s", var);
		assertEvalDobl(EvalException.EXP_NOT_NUMBER, "f.s", var);
		assertEvalLong(EvalException.EXP_NOT_DEF_OBJ, "n.s", var);
		assertEvalLong(EvalException.EXP_NOT_FLD_VALUE, "f.null", var);
		assertEvalLong(EvalException.EXP_NOT_DEF_OBJ, "n.s=1", var);
		assertEvalLong(EvalException.EXP_NOT_LET_FIELD, "f.null=1", var);
	}

	protected void assertEvalLong(int expected, String str, Map var) {
		Expression exp = rule.parse(str);
		if (var != null)
			exp.setVariable(new MapVariable(var));
		assertEvalLong(expected, exp.ae);
	}

	protected void assertEvalLong(int expected, AbstractExpression exp) {
		try {
			System.out.println(exp.evalLong());
		} catch (EvalException e) {
			System.out.println(e);
			assertEquals(expected, e.getErrorCode());
			return;
		}
		fail(exp.toString());
	}

	protected void assertEvalDobl(int expected, String str, Map var) {
		Expression exp = rule.parse(str);
		if (var != null)
			exp.setVariable(new MapVariable(var));
		assertEvalDobl(expected, exp.ae);
	}

	protected void assertEvalDobl(int expected, AbstractExpression exp) {
		try {
			System.out.println(exp.evalDouble());
		} catch (EvalException e) {
			System.out.println(e);
			assertEquals(expected, e.getErrorCode());
			return;
		}
		fail(exp.toString());
	}

	protected void assertEvalObje(int expected, String str, Map var) {
		Expression exp = rule.parse(str);
		if (var != null)
			exp.setVariable(new MapVariable(var));
		assertEvalObje(expected, exp.ae);
	}

	protected void assertEvalObje(int expected, AbstractExpression exp) {
		try {
			System.out.println(exp.evalObject());
		} catch (EvalException e) {
			System.out.println(e);
			assertEquals(expected, e.getErrorCode());
			return;
		}
		fail(exp.toString());
	}

	public void testMessage() {
		assertMessage("Zq̕@G[łB", EvalException.PARSE_INVALID_OP, null);
		assertMessage("Zqu)v݂܂B", EvalException.PARSE_NOT_FOUND_END_OP,
				new String[] { ")" });
	}

	protected void assertMessage(String expected, int code, String[] opt) {
		EvalException e;
		e = new EvalException(code, null, null, -1, null);
		e.msg_opt = opt;
		assertEquals(expected, e.toString());
		e = new EvalException(code, "P", null, -1, null);
		e.msg_opt = opt;
		assertEquals(expected + " word=uPv", e.toString());
		e = new EvalException(code, null, "", -1, null);
		e.msg_opt = opt;
		assertEquals(expected + " string=uv", e.toString());
		e = new EvalException(code, null, null, 0, null);
		e.msg_opt = opt;
		assertEquals(expected + " pos=0", e.toString());
		e = new EvalException(code, null, null, -1, new NullPointerException());
		e.msg_opt = opt;
		assertEquals(expected + " cause by java.lang.NullPointerException", e
				.toString());
		e = new EvalException(code, "", "", -1, null);
		e.msg_opt = opt;
		assertEquals(expected + " string=uv", e.toString());
		e = new EvalException(code, "P", "", 1, null);
		e.msg_opt = opt;
		assertEquals(expected + " word=uPv pos=1 string=uv", e.toString());

		assertEquals(String.valueOf(code), e.toString("%c"));
	}
}

class TestLex extends Lex {

	protected TestLex(String str, List[] lists, AbstractExpression paren,
			ShareExpValue exp) {
		super(str, lists, paren, exp);
	}

	public void check() {
		super.check();
		if (string.equals("^Cv"))
			type = TYPE_ERR;
	}
}

class TestLexFactory extends LexFactory {

	public Lex create(String str, List[] opeList, ShareRuleValue share,
			ShareExpValue exp) {
		return new TestLex(str, opeList, share.paren, exp);
	}
}

class TestRuleFactory extends ExpRuleFactory {

	protected LexFactory getLexFactory() {
		if (defaultLexFactory == null) {
			defaultLexFactory = new TestLexFactory();
		}
		return defaultLexFactory;
	}
}